from collections import namedtuple

import numpy as np
from scipy.optimize import curve_fit

from astropy.modeling import models, fitting
from astropy.stats import sigma_clip
from astropy.table import MaskedColumn
from astropy.coordinates import SkyCoord, match_coordinates_sky
from astropy import units as u

from astroquery.vizier import Vizier


__all__ = [
    'f', 'get_cat', 'opts_to_str', 'calc_residual',
    'filter_transform',
    'calculate_transform_coefficients',
    'transform_magnitudes',
    'transform_to_catalog',
]


def f(X, a, b, c, d, z):
    """
    Calculate the calibrated magnitudes from the instrumental magnitudes and colors.

    Parameters
    ----------

    X : tuple of numpy.ndarray
        The first element is an array of instrumental magnitudes,
        the second is an array of colors.

    a, b, c, d, z : float
        Parameters of the fit.

    Returns
    -------
    `numpy.ndarray`
        Array of calibrated magnitudes.
    """
    mag_inst, color = X

    return a * mag_inst + b * mag_inst ** 2 + c * color + d * color**2 + z


def get_cat(image):
    """
    Get the APASS catalog entries within 1 degree of first object in
    Astropy table 'image'.

    Parameters
    ----------

    image : `astropy.table.Table`
        Table containing the image information. Must have columns
        ``RA`` and ``Dec``.

    Returns
    -------

    `astropy.table.Table`
        Table containing the APASS catalog entries within 1 degree of
        first object in Astropy table.
    """
    our_coords = SkyCoord(image['RA'], image['Dec'], unit='degree')
    # Get catalog via cone search
    Vizier.ROW_LIMIT = -1  # Set row_limit to have no limit
    desired_catalog = 'II/336/apass9'
    a_star = our_coords[0]
    rad = 1 * u.degree
    cat = Vizier.query_region(a_star, radius=rad, catalog=desired_catalog)
    cat = cat[0]
    cat_coords =  SkyCoord(cat['RAJ2000'], cat['DEJ2000'])
    return cat, cat_coords


def opts_to_str(opts):
    """
    Convert the options from a fit to a string.

    Parameters
    ----------

    opts : tuple of float
        Options from a fit.

    Returns
    -------

    str
        String representation of the options.
    """
    opt_names = ['a', 'b', 'c', 'd', 'z']
    names = []
    for name, value in zip(opt_names, opts):
        names.append(f'{name}={value:.4f}')
    return ', '.join(names)


def calc_residual(new_cal, catalog):
    """
    Calculate the standard deviations of the residuals between
    the new calibrated magnitude and the catalog magnitude.

    Parameters
    ----------

    new_cal : `numpy.ndarray`
        New calibrated magnitudes.

    catalog : `numpy.ndarray`
        Catalog magnitudes.

    Returns
    -------

    float
        Standard deviation of the residual.

    """
    resid = new_cal - catalog
    return resid.std()


def filter_transform(mag_data, output_filter,
                     g=None, r=None, i=None,
                     transform=None):
    """
    Transform SDSS magnitudes to BVRI using either the transforms from
    Jester et al or Ivezic et al.

    Parameters
    ----------

    mag_data : `astropy.table.Table`
        Table containing ``g``, ``r`` and ``i`` magnitudes (or at least)
        those required to transform to the desired output filter.

    output_filter : 'B', 'V', 'R' or 'I'
        Filter for which magnitude should be calculated. Note that
        *case matters* here.

    g, r, i : str
        Name of column in table for that magnitude.

    transform : 'jester' or 'ivezic'
        Transform equations to use.

    Returns
    -------

    `astropy.table.Column`
        Output transformed magnitudes as a table column

    Notes
    -----

    The transforms implemented in this function are taken from:

    Jester, et al, *The Sloan Digital Sky Survey View of the Palomar-Green Bright Quasar Survey*, AJ 130, p. 873 (2005)
    http://iopscience.iop.org/article/10.1086/432466/meta

    Ivezić et al, *A Comparison of SDSS Standard Star Catalog for Stripe 82 with Stetson's Photometric Standards*,
    The Future Of Photometric, Spectrophotometric And Polarimetric Standardization, ASP Conference Series 364, p. 165 (2007)
    http://aspbooks.org/custom/publications/paper/364-0165.html

    """
    supported_transforms = ['jester', 'ivezic']
    if transform not in supported_transforms:
        raise ValueError('Transform {} is not known. Must be one of '
                         '{}'.format(transform, supported_transforms))
    transform_ivezic = {
        'B': [0.2628, -0.7952, 1.0544, 0.0268],
        'V': [0.0688, -0.2056, -0.3838, -0.0534],
        'R': [-0.0107, 0.0050, -0.2689, -0.1540],
        'I': [-0.0307, 0.1163, -0.3341, -0.3584]
    }
    base_mag_ivezic = {
        'B': g,
        'V': g,
        'R': r,
        'I': i
    }
    # For jester, using the transform for "all stars with Rc-Ic < 1.15"
    # from
    # http://www.sdss3.org/dr8/algorithms/sdssUBVRITransform.php#Jester2005
    jester_transforms = {
        'B': [1.39, -0.39, 0, 0.21],
        'V': [0.41, 0.59, 0, -0.01],
        'R': [0.41, -0.5, 1.09, -0.23],
        'I': [0.41, -1.5, 2.09, -0.44]
    }

    if output_filter not in base_mag_ivezic.keys():
        raise ValueError('the desired filter must be a string R B V or I')

    if transform == 'ivezic':
        if output_filter == 'R' or output_filter == 'I':
            # This will throw a KeyError if the column is missing
            c = mag_data[r] - mag_data[i]

        if output_filter == 'B' or output_filter == 'V':
            # This will throw a KeyError if the column is missing
            c = mag_data[g] - mag_data[r]

        transform_poly = np.poly1d(transform_ivezic[output_filter])
        out_mag = transform_poly(c) + \
            mag_data[base_mag_ivezic[output_filter]]
        # poly1d  ignores masks. Add masks back in here if necessary.
        try:
            input_mask = c.mask
        except AttributeError:
            pass
        else:
            out_mag = np.ma.array(out_mag, mask=input_mask)
    elif transform == 'jester':
        coeff = jester_transforms[output_filter]
        out_mag = (coeff[0] * mag_data[g] + coeff[1] * mag_data[r] +
                   coeff[2] * mag_data[i] + coeff[3])

    out_mag.name = '{}_mag'.format(output_filter)
    out_mag.description = ('{}-band magnitude transformed '
                           'from gri'.format(output_filter))
    return out_mag


def calculate_transform_coefficients(input_mag, catalog_mag, color,
                                     input_mag_error=None,
                                     catalog_mag_error=None,
                                     faintest_mag=None,
                                     order=1,
                                     sigma=2.0,
                                     gain=None,
                                     ):
    """
    Calculate linear transform coefficients from input magnitudes to catalog
    magnitudes.

    Parameters
    ----------

    input_mag : `numpy.ndarray` or `astropy.table.Column`
        Input magnitudes; for example, instrumental magnitudes.

    catalog_mag : `numpy.ndarray` or `astropy.table.Column`
        Catalog (or reference) magnitudes; the magnitudes to which the
        input_mag will eventually be transformed.

    color : `numpy.ndarray` or `astropy.table.Column`
        Colors to use in determining transform coefficients.

    input_mag_error : `numpy.ndarray` or `astropy.table.Column`, optional
        Error in input magnitudes. Default is zero.

    catalog_mag_error : `numpy.ndarray` or `astropy.table.Column`, optional
        Error in catalog magnitudes. Default is zero.

    faintest_mag_for_transform : float, optional
        If this is not ``None``, the magnitude of the faintest catalog stars
        to use in computing transform coefficients.

    order : int, optional
        Order of the polynomial fit to use in correcting for color.

    sigma : float, optional
        Value of sigma to use to reject outliers while fitting using
        sigma clipping.

    gain : float, optional
        If not ``None``, adjust the instrumental magnitude by
        -2.5 * log10(gain), i.e. gain correct the magnitude.

    verbose : bool, optional
        If ``True``, print some diagnostic information.

    extended_output : bool, optional
        If ``True``, return additional information.

    Returns
    -------

    filtered_data : `numpy.ma.core.MaskedArray`
        The data, with the mask set ``True`` for the data that was *omitted*
        from the fit.

    model : `astropy.modeling.FittableModel`
        Entries in the model are the coefficients in the fit made to the
        data. Since the model is always a polynomial, these are terms in
        a polynomial in the order of ascending power. In other words, the
        coefficient ``ci`` is the coefficient of the term ``x**i``.

    If ``extended_output=True``, then also return:

    fit_input : tuple
        A tuple of color, magnitude for only the stars brighter than
        ``faintest_mag_for_transform``. These are input to the sigma-clipping
        fitter.

    used_in_fit : tuple
        A tuple of color, magnitude for only the stars brighter than
        ``faintest_mag_for_transform`` that were not sigma-cliped out.

    Notes
    -----

    This function has some pretty serious limitations right now:

    + Errors in the independent variable are ignored.
    + Outliers are rejected using a modified loss function (Huber loss) that
      cannot be modified.
    + No errors are estimated in the calculated transformation coefficients.

    And there is all the stuff that is not listed here...
    """

    if input_mag_error is None:
        input_mag_error = np.zeros_like(input_mag)
    if catalog_mag_error is None:
        catalog_mag_error = np.zeros_like(catalog_mag)

    if gain is None:
        gain = 1.0

    # Independent variable is the color, dependent variable is the
    # difference between the measured (input) magnitude and the catalog
    # magnitude.

    mag_diff = catalog_mag - (input_mag - 2.5 * np.log10(gain))

    # The error is the errors of those combined in quadrature.
    combined_error = np.sqrt(input_mag_error**2 + catalog_mag_error**2)

    # If both errors are zero then the error is omitted.
    if (combined_error == 0).all():
        dy = None
    else:
        dy = combined_error

    g_init = models.Polynomial1D(order)
    fit = fitting.LinearLSQFitter()
    or_fit = fitting.FittingWithOutlierRemoval(fit, sigma_clip,
                                               niter=2, sigma=sigma)

    if faintest_mag is not None:
        bright = catalog_mag < faintest_mag
        try:
            bright = bright.filled(False)
        except AttributeError:
            # Might not have had a masked array...
            pass
    else:
        bright = np.ones_like(mag_diff, dtype='bool')

    bright_index = np.nonzero(bright)

    # get fitted model and filtered data
    or_fitted_model, filtered_data_mask = or_fit(g_init,
                                                 color[bright],
                                                 mag_diff[bright])

    # Restore the filtered_data to the same size as the input
    # magnitudes. Unmasked values were included in the fit,
    # masked were not, either because they were too faint
    # or because they were sigma clipped out.
    restored_mask = np.zeros_like(mag_diff, dtype='bool')
    restored_mask[bright_index] = filtered_data_mask
    restored_mask[~bright] = True

    restored_filtered = MaskedColumn(mag_diff.copy())
    restored_filtered.mask = restored_mask

    return (restored_filtered, or_fitted_model)


def transform_magnitudes(input_mags, catalog,
                         transform_catalog,
                         input_mag_colum='mag_inst_r',
                         catalog_mag_column='r_mag',
                         catalog_color_column='B-V',
                         faintest_mag_for_transform=14,
                         sigma=2,
                         order=1,
                         gain=None):
    """
    Calculate catalog magnitudes and transform coefficients
    from instrumental magnitudes.

    Parameters
    ----------

    input_mags : `astropy.table.Table`
        Table which contains a column with instrumental magnitudes, i.e.
        -2.5 * log10(net_counts / exposure_time).

    catalog : `astropy.table.Table`
        Table containing reference catalog of magnitudes and colors.

    transform_catalog : `astropy.table.Table`
        Table containing the reference catalog of magnitudes and colors
        to use in determining the transform coefficients. Can be the
        same table as ``catalog`` if desired.

    input_mag_column : str, optional
        Name of the column in ``input_mags`` with the magnitudes to be
        transformed.

    catalog_mag_column : str, optional
        Name of the column in ``catalog`` with the reference magnitude.

    catalog_color_column : str, optional
        Name of the column in ``catalog`` with color for each star in the
        catalog.

    faintest_mag_for_transform : float, optional
        If this is not ``None``, the magnitude of the faintest catalog stars
        to use in computing transform coefficients.

    sigma : float, optional
        Number of standard deviations to use in rejecting outliers when fitting
        using sigma clipping.

    order : int, optional
        Order of the polynomial to use in fitting magnitude difference/color
        relationship.

    gain : float, optional
        If not ``None``, adjust the instrumental magnitude by
        -2.5 * log10(gain), i.e. gain correct the magnitude.

    Returns
    -------

    our_cat_mags : `astropy.table.Column`
        The calculated catalog magnitudes for the stars in ``input_mags``.

    good_match_all : `numpy.ndarray`
        Boolean array indicating which stars in ``input_mags`` have a match
        in the catalog.

    transforms : `astropy.modeling.FittableModel`
        The coefficients of the transform. The coefficients are in the order
        of ascending power, i.e. the coefficient ``ci`` is the coefficient
        of the term ``x**i``.  Warning: This returns a namedtuple if the fit
        fails.
    """
    catalog_all_coords = SkyCoord(catalog['RAJ2000'],
                                  catalog['DEJ2000'],
                                  unit='deg')

    transform_catalog_coords = SkyCoord(transform_catalog['RAJ2000'],
                                        transform_catalog['DEJ2000'],
                                        unit='deg')
    input_coords = SkyCoord(input_mags['RA'], input_mags['Dec'])

    transform_catalog_index, d2d, _ = \
        match_coordinates_sky(input_coords, transform_catalog_coords)

    # create a boolean of all of the matches that have a discrepancy of less
    # than 5 arcseconds
    good_match_for_transform = d2d < 2 * u.arcsecond

    catalog_index, d2d, _ = match_coordinates_sky(input_coords,
                                                  catalog_all_coords)

    good_match_all = d2d < 5 * u.arcsecond

    catalog_all_indexes = catalog_index[good_match_all]

    input_match_mags = input_mags[input_mag_colum][good_match_for_transform]

    catalog_match_indexes = transform_catalog_index[good_match_for_transform]

    catalog_match_mags = \
        transform_catalog[catalog_mag_column][catalog_match_indexes]
    catalog_match_color = \
        transform_catalog[catalog_color_column][catalog_match_indexes]

    good_mags = ~np.isnan(input_match_mags)

    input_match_mags = input_match_mags[good_mags]
    catalog_match_mags = catalog_match_mags[good_mags]
    catalog_match_color = catalog_match_color[good_mags]

    try:
        matched_data, transforms = \
            calculate_transform_coefficients(input_match_mags,
                                             catalog_match_mags,
                                             catalog_match_color,
                                             sigma=sigma,
                                             faintest_mag=faintest_mag_for_transform,
                                             order=order,
                                             gain=gain)
    except np.linalg.LinAlgError as e:
        print('Danger! LinAlgError: {}'.format(str(e)))
        Transform = namedtuple('Transform', ['parameters'])
        transforms = Transform(parameters=(np.nan,) * (order + 1))

    our_cat_mags = (input_mags[input_mag_colum][good_match_all] +
                    transforms(catalog[catalog_color_column][catalog_all_indexes]))

    return our_cat_mags, good_match_all, transforms


def transform_to_catalog(observed_mags_grouped, obs_mag_col, obs_filter,
                         obs_error_column=None,
                         cat_filter='r_mag', cat_color=('r_mag', 'i_mag'),
                         a_delta=0.5, a_cen=0, b_delta=1e-6, c_delta=0.5, d_delta=1e-6,
                         zero_point_range=(18, 22),
                         in_place=True, fit_diff=True,
                         verbose=True):
    """
    Transform a set of intrumental magnitudes to a standard system using either
    instrumental colors or catalog colors.

    Parameters
    ----------

    observed_magnitudes_grouped : `astropy.table.Table`
        An astropy table, grouped by whatever you want that sepearates the data into
        data from just one image. BJD of the center of the observatory is one reasonable choice

    obs_mag_col : str
        Name of the column in `observed_magnitudes_grouped` that contains instrumental
        magnitudes.

    obs_filter : str
        Name of the filter in which observations were done. Should be one of the names at
        https://www.aavso.org/filters

    obs_error_column : str, optional
        Name of the column in `observed_magnitudes_grouped` that contains the error in the magnitude.

    cat_filter : str
        Name of the filter/passband in catalog that should be matched to the instrumental magnitudes.

    cat_color : tuple of two strings
        Names of the two columns in the caatalog that should be used to calculate color. The magnitude
        difference will be calculated in the ortder the fitlers are given. For example, if the value is
        ``('r_mag', 'i_mag')`` then the calculated color will be the ``r_mag`` column minus
        the ``i_mag`` column.

    a_delta, b_delta, c_delta, d_delta : float, optional
        Range allowed in fitting for each of the parameters ``a``, ``b``, ``c``, and ``d``. Use ``1E-6`` to fix a parameter.

    a_cen : float, optional
        Center of range for the fitting parameter ``a``.

    zero_point_range : tuple of float, optional
        Range to which the value of the zero point is restricted in fitting to observed magnitudes.

    in_place : bool, optional
        If ``True``, add the calibrated magnitude to the input table, othewise return a copy.

    fit_diff : bool, optional
        If ``True``, fit the difference between the instrumental and catalog magnitude instead of the
        treating the catalog mag as the dependent variable.

    verbose: bool optional
        If ``True``, print additional output.

    Returns
    -------

    `astropy.table.Table`
        Table containing the calibrated magnitudes and the fit parameters.

    """

    if obs_error_column is None:
        print("are you sure you want to do that? Error weighting is important!")

    fit_bounds_lower = [
        a_cen - a_delta,     # a
        -b_delta,   # b
        -c_delta,   # c
        -d_delta,   # d
        zero_point_range[0],    # z
    ]

    fit_bounds_upper = [
        a_cen + a_delta,     # a
        b_delta,  # b
        c_delta,  # c
        d_delta,  # d
        zero_point_range[1],    # z
    ]

    fit_bounds = (fit_bounds_lower, fit_bounds_upper)
    zero_point_mid = sum(zero_point_range) / 2

    a = []
    b = []
    c = []
    d = []
    z = []

    all_params = [a, b, c, d, z]

    cal_mags = []
    resids = []
    cat_mags = []
    cat_colors = []
    cat = None
    cat_coords = None

    for file, one_image in zip(observed_mags_grouped.groups.keys, observed_mags_grouped.groups):
        our_coords = SkyCoord(one_image['RA'], one_image['Dec'], unit='degree')
        if cat is None or cat_coords is None:
            cat, cat_coords = get_cat(one_image)
            cat['color'] = cat[cat_color[0]] - cat[cat_color[1]]

        cat_idx, d2d, _ = our_coords.match_to_catalog_sky(cat_coords)

        mag_inst = one_image[obs_mag_col]
        cat_mag = cat[cat_filter][cat_idx]
        color = cat['color'][cat_idx]

        # Impose some constraints on what is included in the fit
        good_cat = ~(color.mask | cat_mag.mask) & (d2d.arcsecond < 1)
        good_dat = ((one_image[obs_mag_col] < -3) &
                     (one_image[obs_mag_col] > -20) &
                     ~np.isnan(one_image[obs_mag_col])
                     )

        mag_diff = cat_mag - mag_inst

        good_data = good_dat & (np.abs(mag_diff - np.nanmedian(mag_diff[good_dat & ~mag_diff.mask])) < 1)
        try:
            good_data = good_data & ~one_image[obs_mag_col].mask
        except AttributeError:
            pass

        good = good_cat & good_data

        # Prep for fitting
        init_guess = (a_cen, 0, 0, 0, zero_point_mid)
        X = (mag_inst[good], color[good])
        catm = cat_mag[good]
        if catm.mask.any() or X[1].mask.any() or np.isnan(X[0]).any():
            print("heck")
        if obs_error_column is not None:
            errors = one_image[obs_error_column][good]
        else:
            errors = None

        if fit_diff:
            offset = X[0]
        else:
            offset = 0

        # Do the fit
        popt, pcov = curve_fit(f, X, catm - offset,
                               p0=init_guess, bounds=fit_bounds,
                               sigma=errors)

        # Accumulate the parameters
        for param, value in zip(all_params, popt):
            param.extend([value] * len(one_image))

        # Calculate and accumulate residual
        residual = calc_residual(f(X, *popt) + offset, catm)
        resids.extend([residual] * len(one_image))

        # Calculate calibrated magnitudes and accumulate, settings ones with no catalog match to NaN
        X = (mag_inst, color)
        cal_mag = f(X, *popt)
        if fit_diff:
            cal_mag = cal_mag + X[0]
        bad_match = d2d.arcsecond > 1
        cal_mag[bad_match] = np.nan
        cal_mags.extend(cal_mag)
        cat_mags.extend(cat_mag)
        cat_colors.extend(color)

        # Keep the user entertained....
        print(f'{file[0]} has fit {opts_to_str(popt)} with {residual=:.4f}')

    mag_col_name = obs_mag_col + '_cal'
    if not in_place:
        result = observed_mags_grouped.copy()
    else:
        result = observed_mags_grouped

    result[mag_col_name] = cal_mags
    if obs_error_column is not None:
        result[mag_col_name + '_error'] = (
            (1 + np.asarray(all_params[0])) * result[obs_error_column]
        )
    opt_names = ['a', 'b', 'c', 'd', 'z']

    for name, values in zip(opt_names, all_params):
        result[name] = values

    result['mag_cat'] = cat_mags
    result['color_cat'] = cat_colors
    return result



import re

import numpy as np

from astroquery.vizier import Vizier

from astropy.coordinates import SkyCoord
import astropy.units as units
from astropy.nddata import block_reduce

__all__ = [
    'in_frame',
    'catalog_search',
    'catalog_clean',
    'find_apass_stars',
    'find_known_variables',
    'filter_catalog'
]


def in_frame(frame_wcs, coordinates, padding=0):
    """
    Given a WCS and list of coordinates check whether the coordinates
    are within the frame.

    Parameters
    ----------

    frame_wcs : `astropy.wcs.WCS`
        WCS for the image.
    coordinates : `astropy.coordinates.SkyCoord`
        Coordinate(s) whose position will be checked to see if they are in the
        field of view.
    padding : int, optional
        Coordinates need to be at least this many pixels in from the edge
        of the frame to be considered in the field of view. Default value
        is 0.

    Returns
    -------

    `numpy.ndarray` of bool
        One value for each input coordinate; values are ``True`` if the
        coordinate was in the field of view, ``False`` otherwise.

    """
    x, y = frame_wcs.all_world2pix(coordinates.ra, coordinates.dec, 0)
    in_x = (x >= padding) & (x <= frame_wcs.pixel_shape[0] - padding)
    in_y = (y >= padding) & (y <= frame_wcs.pixel_shape[1] - padding)
    return in_x & in_y


def catalog_search(frame_wcs_or_center, shape, desired_catalog,
                   ra_column='RAJ2000',
                   dec_column='DEJ2000',
                   radius=0.5,
                   clip_by_frame=True,
                   padding=100):
    """
    Return the items from catalog that are within the search radius and
    (optionally) within the field of view of a frame.

    Parameters
    ----------

    frame_wcs : `astropy.wcs.WCS`
        WCS of the image of interest.

    shape : tuple of int
        Shape of the image of interest.

    desired_catalog : str
        Name of the catalog to be searched.

    ra_column : str, optional
        Name of the column in the catalog that contains the RA values.
        Default is 'RAJ2000'.

    dec_column : str, optional
        Name of the column in the catalog that contains the Dec values.
        Default is 'DEJ2000'.

    radius : float, optional
        Radius, in degrees, around which to search. Default is 0.5.

    clip_by_frame : bool, optional
        If ``True``, only return items that are within the field of view
        of the frame. Default is ``True``.

    padding : int, optional
        Coordinates need to be at least this many pixels in from the edge
        of the frame to be considered in the field of view. Default value
        is 100.

    Returns
    -------

    `astropy.table.Table`
        Table of catalog information for stars in the field of view.

    """
    rad = radius * units.deg
    if isinstance(frame_wcs_or_center, SkyCoord):
        # Center was passed in, just use it.
        center = frame_wcs_or_center
        if clip_by_frame:
            raise ValueError('To clip entries by frame you must use '
                             'a WCS as the first argument.')
    else:
        # Find the center of the frame
        center = frame_wcs_or_center.pixel_to_world(shape[1] / 2,
                                                    shape[0] / 2)

    # Get catalog via cone search
    Vizier.ROW_LIMIT = -1  # Set row_limit to have no limit
    cat = Vizier.query_region(center, radius=rad, catalog=desired_catalog)
    # Vizier always returns list even if there is only one element. Grab that
    # element.
    cat = cat[0]
    cat_coords = SkyCoord(ra=cat[ra_column], dec=cat[dec_column])
    if clip_by_frame:
        in_fov = in_frame(frame_wcs_or_center, cat_coords, padding=padding)
    else:
        in_fov = np.ones([len(cat_coords)], dtype=bool)
    return cat[in_fov]


def catalog_clean(catalog, remove_rows_with_mask=True,
                  **other_restrictions):
    """
    Return a catalog with only the rows that meet the criteria specified.

    Parameters
    ----------

    catalog : `astropy.table.Table`
        Table of catalog information. There are no restrictions on the columns.

    remove_rows_with_mask : bool, optional
        If ``True``, remove rows in which one or more of the values is masked.

    other_restrictions: dict, optional
        Key/value pairs in which the key is the name of a column in the
        catalog and the value is the criteria that values in that column
        must satisfy to be kept in the cleaned catalog. The criteria must be
        simple, beginning with a comparison operator and including a value.
        See Examples below.

    Returns
    -------

    `astropy.table.Table`
        Table of catalog information for stars in the field of view.

    """

    comparisons = {
        '<': np.less,
        '=': np.equal,
        '>': np.greater,
        '<=': np.less_equal,
        '>=': np.greater_equal,
        '!=': np.not_equal
    }

    recognized_comparison_ops = '|'.join(comparisons.keys())
    keepers = np.ones([len(catalog)], dtype=bool)

    if remove_rows_with_mask and catalog.masked:
        for c in catalog.columns:
            keepers &= ~catalog[c].mask

    for column, restriction in other_restrictions.items():
        criteria_re = re.compile(r'({})([-+a-zA-Z0-9]+)'.format(recognized_comparison_ops))
        results = criteria_re.match(restriction)
        if not results:
            raise ValueError("Criteria {}{} not "
                             "understood.".format(column, restriction))
        comparison_func = comparisons[results.group(1)]
        comparison_value = results.group(2)
        new_keepers = comparison_func(catalog[column],
                                      float(comparison_value))
        keepers = keepers & new_keepers

    return catalog[keepers]


def find_apass_stars(image_or_center,
                     radius=1,
                     max_mag_error=0.05,
                     max_color_error=0.1):
    """
    Get APASS data from Vizer.

    Parameters
    ----------

    image_or_center : `astropy.nddata.CCDData` or `astropy.coordinates.SkyCoord`
        Either an image with a WCS (from which the RA/Dec will be extracted) or coordinate of
        the center.

    radius : float, optional
        Radius, in degrees, around which to search. Not needed if the first argument is an image.

    max_mag_error : float, optional
        Maximum error in magnitude to allow. Default is 0.05.

    max_color_error : float, optional
        Maximum error in color to allow. Default is 0.1.

    Returns
    -------

    all_apass : `astropy.table.Table`
        Table of all APASS stars in the field of view.

    apass_lower_error : `astropy.table.Table`
        Table of APASS stars in the field of view with errors lower than the specified values.
    """
    if isinstance(image_or_center, SkyCoord):
        # Center was passed in, just use it.
        cen_wcs = image_or_center
        shape = None
    else:
        cen_wcs = image_or_center.wcs
        shape = image_or_center.shape
    # use the catalog_search function to find the APASS stars in the frame of the image read above
    all_apass = catalog_search(cen_wcs, shape, 'II/336/apass9',
                               ra_column='RAJ2000', dec_column='DEJ2000', radius=radius,
                               clip_by_frame=False)

    # Creates a boolean array of the APASS stars that have well defined
    # magnitudes and color.
    apass_lower_error = (all_apass['e_r_mag'] < max_mag_error) & (
        all_apass['e_B-V'] < max_color_error)

    # create new table of APASS stars that meet error restrictions
    apass_lower_error = all_apass[apass_lower_error]

    return all_apass, apass_lower_error


def find_known_variables(image):
    """
    Get any known variable stars in image field from the VSX catalog.

    Parameters
    ----------

    image : `astropy.nddata.CCDData`
        Image with a WCS.

    Returns
    -------

    vsx : `astropy.table.Table`
        Table of known variable stars in the field of view.
    """
    try:
        vsx = catalog_search(image.wcs, image.shape, 'B/vsx/vsx',
                             ra_column='RAJ2000', dec_column='DEJ2000')
    except IndexError:
        raise RuntimeError('No variables found in this field of view '
                           f'centered on {image.wcs}')
    return vsx


def filter_catalog(catalog, **kwd):
    """
    Filter catalog by key/value pairs in arguments and return bool
    area ``True`` where values in catalog meet the criteria.

    Parameters
    ----------

    catalog : `astropy.table.Table`
        Table whose values are to be filtered.

    kwd : key/value pairs, e.g. ``e_r_mag=0.1``
        The key must be the name of a column and the value the
        *upper limit* on the acceptable values.

    Returns
    -------

    `numpy.ndarray` of bool
        One value for each row in the catalog; values are ``True`` if the
        row meets the criteria, ``False`` otherwise.
    """
    good_ones = np.ones(len(catalog), dtype='bool')

    for key, value in kwd.items():
        print(key, value, catalog[key] <= value)
        good_ones &= (catalog[key] <= value)

    return good_ones



# Class to represent a file in AAVSO extended format
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from importlib import resources
import yaml

from astropy.table import Table, Column

__all__ = [
    "AAVSOExtendedFileFormatColumns",
    "AAVSOExtendedFileFormat"
]


class AAVSOExtendedFileFormatColumns(Enum):
    VARIABLE_MAG = "magnitude"
    VARIABLE_MAG_ERROR = "magerr"
    VARIABLE_STAR_ID_COLUMN = "starid"
    AIRMASS = "airmass"
    JD = "date"
    PASSBAND = "filter"
    CHECK_STAR_MAG = "kmag"
    CHECK_STAR_NAME = "kname"
    COMP_STAR_MAG = "cmag"
    COMP_STAR_NAME = "cname"


@dataclass
class AAVSOExtendedFileFormat:
    """
    A class to represent a file in AAVSO extended format.

    Parameters
    ----------
    obscode : str
        The observer's code.

    Attributes
    ----------
    obscode : str
        The observer's code.

    type : str
        The type of observation. Defaults to 'EXTENDED'.

    software : str
        The software used to generate the file. Defaults to 'stellarphot'.

    delim : str
        The delimiter used in the file. Defaults to ','.

    date_format : str
        The format of the date. Defaults to 'JD'.

    date : `astropy.table.Column`
        The date/time of the observation in the format `date_format`.

    obstype : str
        The type of observation. Defaults to 'CCD'.

    magnitude : `astropy.table.Column`
        The magnitude of the observation.

    magerr : `astropy.table.Column`
        The error in the magnitude of the observation.

    filter : `astropy.table.Column`
        The filter/band used for the observation. Must be one of the AAVSO passbands,
        which are listed at https://www.aavso.org/filters.

    kmag : `astropy.table.Column`
        The magnitude of the check star.

    cmag : `astropy.table.Column`
        The magnitude of the comparison star.

    ensemble : bool
        Whether the observation is an ensemble observation. Defaults to False.

    group : int
        The group number of the ensemble observation. Defaults to 0.

    trans : str
        Whether the observation is a transformed observation. Defaults to 'NO'.

    chart : str
        The chart used for the observation. Defaults to ''.

    notes : str
        Notes about the observation. Defaults to ''.

    starid : str
        The AUID of the star. Defaults to ''.

    cname : str
        The AUID of the comparison star. Defaults to ''.

    kname : str
        The AUID of the check star. Defaults to ''.

    airmass : `astropy.table.Column`
        The airmass of the observation.

    mtype : str
        The type of magnitude. Defaults to 'STD'.
    """
    obscode : str
    _type: str = 'EXTENDED'
    _software: str = 'stellarphot'
    delim: str = ","
    date_format: str = 'JD'
    date: Column = field(default_factory=Column)
    obstype: str = 'CCD'
    magnitude: Column = field(default_factory=Column)
    magerr: Column = field(default_factory=Column)
    filter: Column = field(default_factory=Column)
    kmag: Column = field(default_factory=Column)
    cmag: Column = field(default_factory=Column)
    group : int  = ""
    trans : str = "NO"
    chart : str = ""
    notes : str = ""
    starid : str = ""
    cname : str = ""
    kname : str = ""
    airmass : Column = field(default_factory=Column)
    mtype : str = "STD"

    def __post_init__(self):
        self._ensemble = False

    @property
    def ensemble(self):
        return self._ensemble

    @ensemble.setter
    def ensemble(self, value):
        if value:
            self.cmag = Column(["na"] * len(self.magnitude), name="cmag")
            self.cname = Column(["ENSEMBLE"] * len(self.magnitude), name="cname")

        self._ensemble = value

    @property
    def type(self):
        return self._type

    @property
    def software(self):
        return self._software

    def set_data_columns(self, data, column_map, star_id=None):
        """
        Set the data columns from an astropy table.

        Parameters
        ----------

        data : `astropy.table.Table`
            Table containing the data to be written to the file.
        column_map : dict
            Dictionary mapping the column names in the table to the column names
            in the file.The values in the dictionary should be column names from
            `AAVSOExtendedFileFormatColumns`.
        star_id : str, optional
            ID of the star to be written to the file. If not provided, all stars
            in the table will be written to the file.
        """
        if star_id is not None:
            use_data = data[data['id'] == star_id]
        else:
            use_data = data
        for source_column, destination_column in column_map.items():
            if (self.ensemble and
                (destination_column == AAVSOExtendedFileFormatColumns.COMP_STAR_MAG or
                 destination_column == AAVSOExtendedFileFormatColumns.COMP_STAR_NAME)):
                raise ValueError("Cannot set comparison star magnitude or name for ensemble observation")

            try:
                setattr(self, destination_column.value, use_data[source_column])
            except AttributeError:
                raise AttributeError(f"Column {destination_column.value} not allowed in AAVSO extended format")
            except KeyError:
                raise KeyError(f"Column {source_column} not found in data")

    def set_column_from_single_value(self, value, column):
        setattr(self, column, Column([value] * len(self.variable_mag), name=column))

    def to_table(self):
        """
        Convert the file to an astropy table.

        Returns
        -------
        table : `astropy.table.Table`
            The table representation of the file. Note that there is important information
            in the table's meta attribute.
        """
        # Make a table
        table_description = resources.read_text('stellarphot.io', 'aavso_submission_schema.yml')
        table_structure = yaml.safe_load(table_description)

        table_dict = {}
        length = 0

        # Begin setting up table structure, with column names as keys
        # and the column data as values. If the "column" is really
        # a single item, then make a single entry as the value.
        # It will be turned into a column eventually.
        for key in table_structure['data'].keys():
            # woot! my first walrus operator!
            if isinstance((item := getattr(self, key.lower())), Column):
                # We have a column, but it might not have been set.
                if len(item) == 0:
                    item = "na"
                table_dict[key] = item
                # Keep track of the longest column
                length = len(item) if len(item) > length else length
            else:
                table_dict[key] = item if len(item) > 0 else "na"

        # Fix up anything that is supposed to be a column and isn't yet.
        for k, v in table_dict.items():
            if not isinstance(v, Column) or len(v) != length:
                # If the length is not correct assume we have a single value
                # and make a column out of it.
                table_dict[k] = Column([v] * length, name=k)

        table = Table(table_dict)
        # All of the AAVSO "comment" type entries go in meta.
        table.meta['comments'] = []
        for key in table_structure['comments'].keys():
            if key == "DATE":
                value = self.date_format
            else:
                value = getattr(self, key.lower())
            table.meta['comments'].append(f"{key}={value}")

        # Specify formats for the columns that have formats
        for k, v in table_structure['data'].items():
            length_limit = v.get('limit', None)
            if length_limit:
                if v['type'] == 'float':
                    # For float, the width in the format specifier is the _minimum_
                    # width of the field, not the maximum. Since, most of the time,
                    # any width issues will arise because some rounding is needed,
                    # use the width to calculate the place to which to round the
                    # number, EXCEPT in the case of time. There, use all the digits and
                    # hope for the best.
                    if k != 'DATE':
                        max_precision = length_limit - 3
                        precision = min(max_precision, 6)
                    else:
                        precision = 8

                    # Check whether table column is all "na" -- if that is the case, don't
                    # set a format.
                    if not all(table[k] == "na"):
                        table[k].info.format = f"{length_limit}.{precision}f"
                elif v['type'] == 'str':
                    # For a string, the "precision" in the format is the maximum
                    # size of the field.
                    if k == 'STARID':
                        print(table[k])
                    table[k].info.format = f".{length_limit}s"
                else:
                    raise ValueError(f"Unknown type {v['type']}")
        return table

    def write(self, filename, overwrite=False):
        """
        Write the file to disk.

        Parameters
        ----------
        filename : str, pathlib.Path
            The file to write to.
        """
        table = self.to_table()
        p = Path(filename)
        table.write(p,
                    delimiter=self.delim,
                    comment="#",
                    overwrite=overwrite)



#!/usr/bin/env python

import sys
from setuptools import setup

TEST_HELP = """
Note: running tests is no longer done using 'python setup.py test'. Instead
you will need to run:

    tox -e test

If you don't already have tox installed, you can install it with:

    pip install tox

If you only want to run part of the test suite, you can also use pytest
directly with::

    pip install -e .
    pytest

For more information, see:

  http://docs.astropy.org/en/latest/development/testguide.html#running-tests
"""

if 'test' in sys.argv:
    print(TEST_HELP)
    sys.exit(1)

DOCS_HELP = """
Note: building the documentation is no longer done using
'python setup.py build_docs'. Instead you will need to run:

    tox -e build_docs

If you don't already have tox installed, you can install it with:

    pip install tox

For more information, see:

  http://docs.astropy.org/en/latest/install.html#builddocs
"""

if 'build_docs' in sys.argv or 'build_sphinx' in sys.argv:
    print(DOCS_HELP)
    sys.exit(1)

setup(use_scm_version=True)



from pathlib import Path
import warnings

from astropy.io import fits
from astropy.nddata import CCDData

from ipyfilechooser import FileChooser


__all__ = ['FitsOpener']


class FitsOpener:
    """ A class to open FITS files and display them in an `astrowidgets.ImageWidget`.

    Attributes
    ----------

    ccd : `astropy.nddata.CCDData`
        The selected FITS file as a CCDData object.

    file_chooser : `ipyfilechooser.FileChooser`
        The actual FileChooser widget.

    header : dict
        The header of the selected FITS file.

    path : `pathlib.Path`
        The path to the selected FITS file.

    register_callback : function, optional
        A function that takes one argument. This function will be called when the
        selected file changes.
    """
    def __init__(self, title="Choose an image", filter_pattern=None, **kwargs):
        """
        Initializes an instance of the FitsOpener class, which is a wrapper around
        the `ipyfilechooser.FileChooser` widget that (if no `filter_pattern` is given)
        defaults to showing only FITS files.

        Parameters
        ----------

        title : str, optional
            The title of the FileChooser widget. The default is "Choose an image".

        filter_pattern : str, optional
            The filter pattern to use for the FileChooser widget. The default is None,
        """
        self._fc = FileChooser(title=title, **kwargs)
        if not filter_pattern:
            self._fc.filter_pattern = ['*.fit*', '*.fit*.[bg]z']
        else:
            self._fc.filter_pattern = filter_pattern

        self._header = {}
        self._selected_cache = self._fc.selected
        self.object = ""
        self.register_callback(lambda x: None)

    @property
    def file_chooser(self):
        """
        The actual FileChooser widget.
        """
        return self._fc

    @property
    def header(self):
        self._set_header()
        return self._header

    @property
    def ccd(self):
        """
        Return image as CCDData object
        """
        return CCDData.read(self.path)

    @property
    def path(self):
        return Path(self._fc.selected)

    def _set_header(self):
        if not self._header or self._fc.selected != self._selected_cache:
            self._selected_cache = self._fc.selected
            try:
                self._header = fits.getheader(self.path)
            except OSError:
                return

        try:
            self.object = self._header['object']
        except KeyError:
            pass

    def register_callback(self, callable):
        """
        Register a callback that is called when the value changes. This is
        the alternative to observing a value, which is not implemented for
        some reason. Only one callback function is allowed.

        This wraps the user-supplied callable to also update the header.

        Parameters
        ----------

        callable : function
            A function that takes one argument.
        """
        def wrap_call(change):
            self._set_header()
            callable(change)

        self._fc.register_callback(wrap_call)

    def load_in_image_widget(self, image_widget):
        """
        Load the selected image into an `astrowidgets.ImageWidget`, doing the best
        that can be done to suppress warnings.

        Parameters
        ----------

        image_widget : `astrowidgets.ImageWidget`
            The widget into which to load the image.
        """
        with warnings.catch_warnings():
            image_widget.load_fits(str(self.path))

    def set_file(self, file, directory=None):
        """
        Set the selected file of the `ipyfilechooser.FileChooser` to be the input file.

        Parameters
        ----------

        file : Path-like
            The file to be set as selected. Can be a string or a pathlib.Path.
            Cannot contain any directory information.

        directory : Path-like, optional
            Directory the file is in. The default value is the current directory of the
            `ipyfilechooser.FileChooser`.
        """
        if directory is None:
            directory = self._fc.selected_path

        self._fc.reset(directory, file)
        self._fc._apply_selection()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# Packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *   # noqa
# ----------------------------------------------------------------------------

# Enforce Python version check during package import.
# This is the same check as the one at the top of setup.py
import sys
from distutils.version import LooseVersion

__minimum_python_version__ = "3.7"

__all__ = []


class UnsupportedPythonError(Exception):
    """
    Wraps an exception to raise if a user tries to use stellarphot
    with an unsupported version of Python.
    """
    pass


if LooseVersion(sys.version) < LooseVersion(__minimum_python_version__):
    raise UnsupportedPythonError("stellarphot does not support Python < {}"
                                 .format(__minimum_python_version__))

from .core import *
from .source_detection import *
from .photometry import *



from astropy import units as u

__all__ = ['Camera']


class Camera:
    """
    A class to represent a CCD-based camera.

    Attributes
    ----------

    gain : `astropy.quantity.Quantity`
        The gain of the camera in units such that the unit of the product
        `gain` times the image data matches the unit of the `read_noise`.

    read_noise : `astropy.quantity.Quantity`
        The read noise of the camera in units of electrons.

    dark_current : `astropy.quantity.Quantity`
        The dark current of the camera in units such that, when multiplied
        by exposure time, the unit matches the unit of the `read_noise`.

    Notes
    -----
    The gain, read noise, and dark current are all assumed to be constant
    across the entire CCD.

    Examples
    --------
    >>> from astropy import units as u
    >>> from stellarphot import Camera
    >>> camera = Camera(gain=1.0 * u.electron / u.adu,
    ...                 read_noise=1.0 * u.electron,
    ...                 dark_current=0.01 * u.electron / u.second)
    >>> camera.gain
    <Quantity 1. electron / adu>
    >>> camera.read_noise
    <Quantity 1. electron>
    >>> camera.dark_current
    <Quantity 0.01 electron / s>

    """
    def __init__(self, gain=1.0 * u.electron / u.adu,
                 read_noise=1.0 * u.electron,
                 dark_current=0.01 * u.electron / u.second):
        """
        Initializes a Camera object instance.

        Parameters
        ----------
        gain : `astropy.quantity.Quantity`
            The gain of the camera in units such that the unit of the product `gain`
            times the image data matches the unit of the `read_noise`.

        read_noise : `astropy.quantity.Quantity`
            The read noise of the camera in units of electrons.

        dark_current : `astropy.quantity.Quantity`
            The dark current of the camera in units such that, when multiplied by
            exposure time, the unit matches the unit of the `read_noise`.
        """
        super().__init__()
        self.gain = gain
        self.read_noise = read_noise
        self.dark_current = dark_current
